import numpy as np
from numpy import linalg as la

from . import CheckerboardLattice


class ExtendedBasisLattice(CheckerboardLattice.CheckerboardLattice):
    def __init__(
        self, Vtot, Ratio, n_plane_waves, Frequency, K0, nFloquetBasis, Ellipse_Factor=1
    ):

        super().__init__(Vtot, Ratio, n_plane_waves)
        self.Frequency = Frequency
        # Coupling strength for circles needs to be adjusted by 1+i since there is a relative phase between x and y
        self.CouplingStrength = K0 * self.Er / np.pi
        self.nFloquetBasis = nFloquetBasis
        self.n_freq = 2 * nFloquetBasis + 1
        self.Ellipse_Factor = Ellipse_Factor

    def DiagonalizeExtendedBasis_qx(self, q_points, n_to_keep=25, qy=0):

        self.qx = np.linspace(-1, 1, q_points)
        self.ExtendedEnergies = np.zeros((q_points, n_to_keep))

        for loop, kx in enumerate(self.qx):
            self.Hd = (
                -2 * self.n_plane_waves
                + np.mod(self.idx - 1, self.plane_wave_dim_1d)
                + np.floor((self.idx - 1) / self.plane_wave_dim_1d)
                + kx
            ) ** 2 + (
                np.mod(self.idx - 1, self.plane_wave_dim_1d)
                - np.floor((self.idx - 1) / self.plane_wave_dim_1d)
                + qy
            ) ** 2

            self.H0 = self.Hod + (self.Er * np.diag(self.Hd))
            self.SetUpExtendedBasis(kx, qy)

            self.ExtendedEnergies[loop, :] = np.sort(la.eigvalsh(self.H_extended))[
                :n_to_keep
            ]

    def DiagonalizeExtendedBasis_GMKG(self, n_per_line, n_to_keep=25):
        """
        Diagonalize in the extened basis along the Gamma->M->K->Gamma Line.  In coordinates, it is
        (0,0)->(1,0) ->(0,1)->(0,0)

        Parameters:
        n_per_line: int
          Number of points to consider along each line section.  Total number of diagonalizations will be this x3.
        """

        # (0,0) -> (1,0)
        qx1 = np.linspace(0, 1, n_per_line)
        qy1 = 0 * qx1

        # (1,0) -> (1/2,1/2)
        qx2 = np.linspace(1, 1 / 2, n_per_line // 2)
        qy2 = np.linspace(0, 1 / 2, n_per_line // 2)

        # (1/2,1/2) -> (0,0)
        qy3 = np.linspace(1 / 2, 0, n_per_line // 2)
        qx3 = 1 * qy3

        self.qx = np.concatenate([qx1, qx2, qx3])
        self.qy = np.concatenate([qy1, qy2, qy3])

        self.ExtendedEnergies = np.zeros((len(self.qx), n_to_keep))

        for loop, (kx, ky) in enumerate(zip(self.qx, self.qy)):
            self.ExtendedEnergies[loop, :] = self.DiagonalizePoint(kx, ky, n_to_keep)

    def DiagonalizePoint(self, qx, qy, n_to_keep=25):
        self.Hd = (
            -2 * self.n_plane_waves
            + np.mod(self.idx - 1, self.plane_wave_dim_1d)
            + np.floor((self.idx - 1) / self.plane_wave_dim_1d)
            + qx
        ) ** 2 + (
            np.mod(self.idx - 1, self.plane_wave_dim_1d)
            - np.floor((self.idx - 1) / self.plane_wave_dim_1d)
            + qy
        ) ** 2
        self.H0 = self.Hod + (self.Er * np.diag(self.Hd))
        self.SetUpExtendedBasis(qx, qy)

        return np.sort(la.eigvalsh(self.H_extended))[:n_to_keep]

    def SetUpExtendedBasis(self, qx, qy):
        """
        Generate the Hamiltonian in the extended basis.  Given the static Hamiltonian H0, this matrix is

        | H0 - w  | H1  | 0     | ... |
        |______________________________
        | H-1     | H0  | H1    | ... |
        |_____________________________
        |0        | H-1 | H0 + w| H1..|

        This is a tri-block diagonal matrix where the diagonal blocks are Ho - nw I and the
        off diagonal blocks H1/H-1 are the +/- 1 fourier components.  Since we only consider a
        single frequency, we only have H1 = CouplingStrength*momentum (which will be diaoonal).  The coupling strength
        is technically Shake_x + Shake_y, so for a circular shake, the coupling needs
        to be adjusted by 1j in a direction to account for the phase.
        """

        self.floquet_freq = np.arange(-self.nFloquetBasis, self.nFloquetBasis + 1)

        # Generate the diagonal H0 - nwI.  The full H0 diagonal block will just be
        # np.kron( H0, np.eye(2*nFloquet+1)) which repeats H0 along the diagonal 2*nFloquet+1 times
        self.Hd = np.kron(np.eye(2 * self.nFloquetBasis + 1), self.H0)

        # This will add -nw to the diagonal components
        self.H0_nomega = self.Frequency * np.kron(
            np.diag(self.floquet_freq), np.eye(self.H0.shape[0])
        )

        self.Hdiag = self.Hd + self.H0_nomega

        # Generate H1 and H-1 which will just be coupling_strength*momentum, where momentum will be the q+n*k plane wave component
        # I am making the qy component imaginary to account for circular shaking

        self.CouplingMatrix = np.diag(
            self.CouplingStrength
            * (
                (
                    -2 * self.n_plane_waves
                    + np.mod(self.idx - 1, self.plane_wave_dim_1d)
                    + np.floor((self.idx - 1) / self.plane_wave_dim_1d)
                    + qx
                )
                + (
                    np.mod(self.idx - 1, self.plane_wave_dim_1d)
                    - np.floor((self.idx - 1) / self.plane_wave_dim_1d)
                    + qy
                )
                * 1j
                * self.Ellipse_Factor
            )
        )
        # This is a (2n+1)^2 x (2n+1)^2 block  It needs to be along the +1 and -1 diagonal
        self.H1 = np.kron(
            np.diag(np.ones(2 * self.nFloquetBasis), 1), self.CouplingMatrix
        ) + np.conj(
            np.kron(np.diag(np.ones(2 * self.nFloquetBasis), -1), self.CouplingMatrix)
        )
        self.H_extended = self.Hdiag + self.H1



import numpy as np
from numpy import linalg as la, cos

from . import CheckerboardLattice

class TrotterLattice(CheckerboardLattice.CheckerboardLattice):

    def __init__(self, Vtot, Ratio, n_plane_waves):
        super().__init__(Vtot, Ratio, n_plane_waves)
        self.GenerateOffDiagonal()

    def shake(self, kx, ky, Omega, K0, N_trotter, phase=np.pi / 2, dx=None, dy=None):
        """
        Get Shaken Eigenvalues for given (kx,ky).  Result stored
        in self.Floquet_E
        """
        self.Omega = Omega
        self.K0 = K0
        self.Period_ = 2 * np.pi / self.Omega
        self.NT_ = N_trotter
        self.phase = phase
        self.times_ = np.linspace(0, self.Period_, self.NT_)
        self.dt = self.times_[2] - self.times_[1]
        Amp = 2 / 7.5 * K0 / (0.54 * Omega * 10 ** (-3))

        if dx:
            self.dx = dx
        else:
            self.dx = Amp * np.pi / 0.406
        if dy:
            self.dy = dy
        else:
            self.dy = 1 * self.dx

        self.Hd = (
            -2 * self.n_plane_waves
            + np.mod(self.idx - 1, self.plane_wave_dim_1d)
            + np.floor((self.idx - 1) / self.plane_wave_dim_1d)
            + kx
        ) ** 2 + (
            np.mod(self.idx - 1, self.plane_wave_dim_1d)
            - np.floor((self.idx - 1) / self.plane_wave_dim_1d)
            + ky
        ) ** 2

        self.H = self.Hod + (self.Er * np.diag(self.Hd))

        static_energies, static_vectors = la.eigh(self.H)
        static_vectors = static_vectors[:, np.argsort(static_energies)]
        static_energies = np.sort(static_energies)
        BlochBasis = np.diag(np.exp(-1j * static_energies * self.dt))
        FourierBasis = np.matmul(
            static_vectors, np.matmul(BlochBasis, np.transpose(static_vectors))
        )
        # Trotter Matrix
        TrotterMat = np.eye(self.plane_wave_dim_1d ** 2, dtype=complex)

        for t in self.times_:
            DiagonalTM = np.exp(
                1j
                / 2
                * self.dt
                * self.Omega
                * (
                    self.dx
                    * (cos(self.Omega * t))
                    * (
                        -2 * self.n_plane_waves
                        + np.mod(self.idx - 1, self.plane_wave_dim_1d)
                        + np.floor((self.idx - 1) / self.plane_wave_dim_1d)
                        + kx
                    )
                    + self.dy
                    * (cos(self.Omega * t + self.phase))
                    * (
                        np.mod(self.idx - 1, self.plane_wave_dim_1d)
                        - np.floor((self.idx - 1) / self.plane_wave_dim_1d)
                        + ky
                    )
                )
            )
            dTM = np.diag(DiagonalTM)
            TrotterMat = np.matmul(
                dTM, np.matmul(FourierBasis, np.matmul(dTM, TrotterMat))
            )
            shakeEigs, ShakeVecs = la.eig(TrotterMat)
            vecs_dot_static = np.abs(np.dot(np.transpose(ShakeVecs), static_vectors))
            max_idx = np.argmax(vecs_dot_static, axis=0)
            if max_idx[0] == max_idx[1]:
                max_idx[1] = np.argsort(vecs_dot_static[:, 1])[-2]
            SortedShaking = shakeEigs[max_idx]
            self.Floquet_E = np.real(1j * np.log(SortedShaking)) / self.Period_



import numpy as np
from numpy import linalg as la


class CheckerboardLattice:
    Er = 3478

    def __init__(self, Vtot, Ratio, n_plane_waves):

        self.Vtot = Vtot
        self.Ratio = Ratio
        self.V2D = Vtot * Ratio * self.Er
        self.Vop = Vtot * (1 - Ratio) * self.Er
        self.Vz = self.Vop
        self.n_plane_waves = n_plane_waves
        self.plane_wave_dim_1d = 2 * n_plane_waves + 1
        self.plane_wave_dim_2d = (2 * n_plane_waves + 1) ** 2

        self.couplings = {
            "A": -self.V2D / 8 - self.Vop / 16,
            "B": -self.Vop / 8,
            "C": np.conj(-self.V2D / 8 - self.Vop / 16),
            "D": np.conj(-self.Vop / 8),
        }
        self.GenerateOffDiagonal()
        self.res = self.q0_gaps([(0, 1)])[0]

    def GenerateOffDiagonal(self):
        A = self.couplings["A"]
        B = self.couplings["B"]
        C = self.couplings["C"]
        D = self.couplings["D"]
        dim = self.plane_wave_dim_1d
        idx = np.linspace(1, dim ** 2, dim ** 2)
        self.idx = idx
        C1 = B * np.ones((1, dim ** 2 - 1)) * (np.mod(idx[0 : dim ** 2 - 1], dim) > 0)
        Cm1 = D * np.ones((1, dim ** 2 - 1)) * (np.mod(idx[0 : dim ** 2 - 1], dim) > 0)
        C2 = (
            A
            * np.ones((1, dim ** 2 - (dim + 1)))
            * (np.mod(idx[0 : dim ** 2 - (dim + 1)], dim) > 0)
        )
        Cm2 = (
            C
            * np.ones((1, dim ** 2 - (dim + 1)))
            * (np.mod(idx[0 : dim ** 2 - (dim + 1)], dim) > 0)
        )
        C3 = B * np.ones((1, dim ** 2 - dim))
        Cm3 = D * np.ones((1, dim ** 2 - dim))
        C4 = (
            A
            * np.ones((1, dim ** 2 - (dim - 1)))
            * (np.mod(idx[0 : dim ** 2 - (dim - 1)], dim) > 0)
        )
        Cm4 = (
            C
            * np.ones((1, dim ** 2 - (dim - 1)))
            * (np.mod(idx[0 : dim ** 2 - (dim - 1)] - 1, dim) > 0)
        )
        self.Hod = (
            np.diag((C1[0]), 1)
            + np.diag(Cm1[0], -1)
            + np.diag(C2[0], dim + 1)
            + np.diag(Cm2[0], -dim - 1)
            + np.diag(C3[0], dim)
            + np.diag(Cm3[0], -dim)
            + np.diag(C4[0], dim - 1)
            + np.diag(Cm4[0], -(dim - 1))
        )

    def diagonalize_qx(self, q_points, qy=0):
        self.qx = np.linspace(-1, 1, q_points)
        self.StaticEnergies = np.zeros((q_points, self.plane_wave_dim_2d))
        for loop, kx in enumerate(self.qx):
            self.Hd = (
                -2 * self.n_plane_waves
                + np.mod(self.idx - 1, self.plane_wave_dim_1d)
                + np.floor((self.idx - 1) / self.plane_wave_dim_1d)
                + kx
            ) ** 2 + (
                np.mod(self.idx - 1, self.plane_wave_dim_1d)
                - np.floor((self.idx - 1) / self.plane_wave_dim_1d)
                + qy
            ) ** 2

            self.H = self.Hod + (self.Er * np.diag(self.Hd))
            self.StaticEnergies[loop, :] = np.sort(la.eigvalsh(self.H))

    def diagonalize_q(self, qx, qy, N_energies=2):
        self.Hd = (
            -2 * self.n_plane_waves
            + np.mod(self.idx - 1, self.plane_wave_dim_1d)
            + np.floor((self.idx - 1) / self.plane_wave_dim_1d)
            + qx
        ) ** 2 + (
            np.mod(self.idx - 1, self.plane_wave_dim_1d)
            - np.floor((self.idx - 1) / self.plane_wave_dim_1d)
            + qy
        ) ** 2
        self.H = self.Hod + (self.Er * np.diag(self.Hd))
        return np.sort(la.eigvalsh(self.H))[:N_energies]

    def q0_gaps(self, band_indexes):
        """
        Get q = 0 gap between given band indices.   band_indexes should be a list of tuples
        containing (low band, high band), so band_indexes = (0,1) gives the difference between
        lowest 2 bands
        """
        n_to_get = np.max(band_indexes) + 1
        Energies = self.diagonalize_q(0, 0, n_to_get)
        energies = []
        for transitions in band_indexes:
            energies.append(Energies[transitions[1]] - Energies[transitions[0]])
        return energies



from setuptools import setup

setup(
    name="ShakenCBLattice",
    version="0.1.0",
    description="Package used to perform band structure caluclations on periodically driven Checkerboard optical lattice",
    author="James Maslek",
    author_email="jmaslek11@gmail.com",
    packages=["ShakenCBLattice"],
)
